@@ -163,10 +163,9 @@ import py_cui
 from colorama import Fore, Style, init
 
 from awsrun.argparse import AppendAttributeValuePair
-from awsrun.config import Bool, Choice, Dict, Int, Str, List
+from awsrun.config import Bool, Choice, Dict, Int, List, Str
 from awsrun.runner import RegionalCommand
 
-
 # Lookup attributes supported by AWS CloudTrail
 LOOKUP_ATTRIBUTES = [
     "EventId",
@@ -392,96 +391,164 @@ class CLICommand(RegionalCommand):
         if not self.interactive:
             return
 
-        # User wants the TUI, so let's fire it up! Keep in mind, the Python
-        # library I found to make the TUI is barely alpha, so I've had to work
-        # around many of its limitations that I will submit to the author.
-        if self.vertical:
-            root = _MyCUI(4, 1)
-            user_list = root.add_my_scroll_menu(f"Usernames", 0, 0)
-            event_list = root.add_my_scroll_menu("Events", 1, 0)
-            event_detail = root.add_my_scroll_menu("Event Detail", 2, 0, row_span=2)
+        # Let's fire up the TUI!
 
-        else:  # Default hybrid layout
-            root = _MyCUI(3, 4)
-            user_list = root.add_my_scroll_menu(f"Usernames", 0, 0, column_span=2)
-            event_list = root.add_my_scroll_menu(
-                "Events", 1, 0, row_span=2, column_span=2
-            )
-            event_detail = root.add_my_scroll_menu(
+        # These contain the list of events being displayed by the TUI widgets.
+        # By default, all events are shown for a specific unless the user
+        # applies a filter.
+        events = self.events
+        events_by_user = self.events_by_user
+
+        # This contains the current query expression to filter events. We only
+        # save a reference to this, so we can prepopulate the filter popup box
+        # with the last expression the user used. This makes it easy for a user
+        # to keep tweaking long exressions without having to type it over and
+        # over again.
+        current_filter_expression = ""
+
+        # Our layout consists of three panes: user list, event list, and event
+        # detail.  We support a vertical layout where all three are stacked
+        # vertically in the terminal for those with narrow terminals ...
+        if self.vertical:
+            root = py_cui.PyCUI(4, 1)
+            user_list = root.add_scroll_menu(f"Usernames", 0, 0)
+            event_list = root.add_scroll_menu("Events", 1, 0)
+            event_detail = root.add_scroll_menu("Event Detail", 2, 0, row_span=2)
+
+        # ... but the default layout consists of the user and event list
+        # stacked vertically on the left side of the screen with the event
+        # detail pane taking the full height on the right side.
+        else:
+            root = py_cui.PyCUI(3, 4)
+            user_list = root.add_scroll_menu(f"Usernames", 0, 0, column_span=2)
+            event_list = root.add_scroll_menu("Events", 1, 0, row_span=2, column_span=2)
+            event_detail = root.add_scroll_menu(
                 "Event Detail", 0, 2, row_span=3, column_span=2
             )
 
         root.toggle_unicode_borders()
         root.set_title("CloudTrail Events")
-        root.set_status_bar_text("Press - q - to exit. TAB to move between widgets.")
+        root.set_status_bar_text("Press 'q' to exit. TAB to switch panes.")
 
-        # Override the default behavior and allow TAB to switch between
-        # widgets and activate widgets.
-        focus = cycle(root.widgets.values())
+        # py_cui has an odd navigation mechanism where the TUI is either in
+        # overview mode or focus mode.  In overview mode, users navigate with
+        # arrow keys to navigate between widgets, and then press RETURN to
+        # enter focus mode. Focus mode allows users to interact with the widget
+        # only until the user presses ESC to return to overview mode. I find
+        # this non-intuitive, so let's provide a mappting for TAB to cycle
+        # between widgets and place them in focus mode automatically.
+        focus = cycle(root.get_widgets().values())
 
         def select_next_widget():
             root.move_focus(next(focus))
 
         root.add_key_command(py_cui.keys.KEY_TAB, select_next_widget)
 
-        # Configure our user list
+        # Filter the event list objects in the outer scope that are used by
+        # widgets to display content. This function is called as a result of
+        # the user entering a query expression in the filtering popup.
+        def filter_event_lists(expr):
+            nonlocal events
+            nonlocal events_by_user
+
+            if not expr.strip():
+                events = self.events
+                events_by_user = self.events_by_user
+                return
+
+            events_by_user = defaultdict(list)
+            for user, events in self.events_by_user.items():
+                events = _filter_events(events, expr)
+                if events:
+                    events_by_user[user] = events
+
+            events = []  # List of lists
+            for sublist in self.events:
+                events.append(_filter_events(sublist, expr))
+
+        # Callback used when user presses key to open filtering popup. By
+        # default, py_cui returns the user to overview mode after the popup is
+        # dismissed. We save the widget that was in focus before the popup, so
+        # we can refocus on it afterwards.
+        def show_filter_popup():
+            original_widget = root.get_selected_widget()
+
+            def popup_callback(s):
+                nonlocal current_filter_expression
+                current_filter_expression = s
+                filter_event_lists(s)
+                update_user_list()
+                root.move_focus(original_widget)
+
+            root.show_text_box_popup("Filter expression:", popup_callback)
+            # TODO Workaround until upstream allows setting initial text.
+            root._popup.set_text(current_filter_expression)
+            # TODO Workaround until upstream py_cui fixes bug
+            root._popup._initial_cursor = (
+                root._popup.get_start_position()[0] + root._popup.get_padding()[0] + 2
+            )
+
+        def update_user_list():
+            user_list.clear()
+            user_list.add_item_list(
+                sorted(events_by_user.keys(), key=lambda s: s.lower())
+            )
+            user_list.set_title(f"Usernames ({len(events_by_user.keys())})")
+            update_event_list()
+
         def update_event_list():
             event_list.clear()
-            event_list.add_item_list(self.events_by_user[user_list.get()])
-            event_list.title = f"Events ({len(event_list.get_item_list())})"
+            if user_list.get():
+                event_list.add_item_list(events_by_user[user_list.get()])
+            event_list.set_title(f"Events ({len(event_list.get_item_list())})")
             update_event_detail()
 
+        def update_event_detail():
+            event_detail.clear()
+            if event_list.get():
+                event_detail.add_item_list(event_list.get().to_json().split("\n"))
+
         def unfilter_event_list():
             event_list.clear()
             all_events = sorted(
-                chain(*self.events), reverse=True, key=lambda e: e.event["EventTime"]
+                chain(*events), reverse=True, key=lambda e: e.event["EventTime"]
             )
             event_list.add_item_list(all_events)
-            event_list.title = f"Events ({len(all_events)})"
+            event_list.set_title(f"Events ({len(all_events)})")
             update_event_detail()
 
+        user_list.set_selected_color(self.color)
+        user_list.set_focus_border_color(self.color)
         user_list.add_key_command(py_cui.keys.KEY_TAB, select_next_widget)
         user_list.add_key_command(py_cui.keys.KEY_ENTER, update_event_list)
         user_list.add_key_command(py_cui.keys.KEY_BACKSPACE, unfilter_event_list)
-        user_list.add_key_command(py_cui.keys.KEY_PAGE_UP, user_list.page_up)
-        user_list.add_key_command(py_cui.keys.KEY_PAGE_DOWN, user_list.page_down)
+        user_list.add_key_command(py_cui.keys.KEY_F_LOWER, show_filter_popup)
         user_list.add_key_command(py_cui.keys.KEY_Q_LOWER, lambda: root.stop())
-        user_list.title = f"Usernames ({len(self.events_by_user.keys())})"
         user_list.set_focus_text(
-            "Press - q - to exit. TAB to move between widgets. ENTER to filter events. BACKSPACE to show all."
-        )
-        user_list.set_selected_color(self.color)
-        user_list.add_item_list(
-            sorted(self.events_by_user.keys(), key=lambda s: s.lower())
+            "Press 'q' to exit. TAB to switch panes. RET for user events. BACKSPACE for all events. 'f' to filter."
         )
 
-        # Configure our event list
-        def update_event_detail():
-            event_detail.clear()
-            event_detail.add_item_list(event_list.get().to_json().split("\n"))
-
+        event_list.set_selected_color(self.color)
+        event_list.set_focus_border_color(self.color)
         event_list.add_key_command(py_cui.keys.KEY_TAB, select_next_widget)
         event_list.add_key_command(py_cui.keys.KEY_ENTER, update_event_detail)
-        event_list.add_key_command(py_cui.keys.KEY_PAGE_UP, event_list.page_up)
-        event_list.add_key_command(py_cui.keys.KEY_PAGE_DOWN, event_list.page_down)
+        event_list.add_key_command(py_cui.keys.KEY_F_LOWER, show_filter_popup)
         event_list.add_key_command(py_cui.keys.KEY_Q_LOWER, lambda: root.stop())
         event_list.set_focus_text(
-            "Press - q - to exit. TAB to move between widgets. ENTER to display event detail."
+            "Press 'q' to exit. TAB to switch panes. RET for event detail. 'f' to filter."
         )
-        event_list.set_selected_color(self.color)
-        if user_list.get():
-            update_event_list()
 
-        # Configure our event detail
-        event_detail.add_key_command(py_cui.keys.KEY_TAB, select_next_widget)
-        event_detail.add_key_command(py_cui.keys.KEY_PAGE_UP, event_detail.page_up)
-        event_detail.add_key_command(py_cui.keys.KEY_PAGE_DOWN, event_detail.page_down)
-        event_detail.add_key_command(py_cui.keys.KEY_Q_LOWER, lambda: root.stop())
-        event_detail.set_focus_text("Press - q - to exit. TAB to move between widgets.")
         event_detail.set_selected_color(self.color)
-        if event_list.get():
-            update_event_detail()
+        event_detail.set_focus_border_color(self.color)
+        event_detail.add_key_command(py_cui.keys.KEY_TAB, select_next_widget)
+        event_detail.add_key_command(py_cui.keys.KEY_F_LOWER, show_filter_popup)
+        event_detail.add_key_command(py_cui.keys.KEY_Q_LOWER, lambda: root.stop())
+        event_detail.set_focus_text(
+            "Press 'q' to exit. TAB to switch panes. 'f' to filter."
+        )
 
+        # Load the widgets with data and select the user list.
+        update_user_list()
         select_next_widget()
 
         # Fire up the main event loop
@@ -526,6 +593,15 @@ class _UserIdentityType:
                 return user
         return self._parse_username()
 
+    def contains(self, s):
+        """Return True if the event contains `s`."""
+
+        try:
+            next(_deep_finder(self.ct_event, lambda n: isinstance(n, str) and s in n))
+            return True
+        except StopIteration:
+            return False
+
     def _parse_username(self):
         return self.event.get(
             "Username", self.user_identity.get("userName", "NO_USERNAME")
@@ -551,7 +627,10 @@ class _UserIdentityType:
         src = self.event.get("EventSource", "")
         time = self.event.get("EventTime", "")
         name = self.event.get("EventName", "")
-        return f"{time}  {src:25.25} {name:30.30} {self.username()}"
+        error = self.ct_event.get("errorCode", "")
+        if error:
+            error = f"ERROR: {error}"
+        return f"{time}  {src:25.25} {name:30.30} {self.username()}  {error}"
 
 
 #############################################################################
@@ -624,6 +703,54 @@ def _lookup_events(ct, start, end, attrs=None):
     )
 
 
+def _filter_events(events, query):
+    """Return list of events filtered by query expression.
+
+    Query expression may consist of one or more terms. Terms are matched using
+    logical OR. A term may be prefixed with an optional '-' to exclude events
+    containing the term. Terms are case sensitive and are matched as substrings
+    in a CloudTrail event including both keys and values.
+
+    For example, to search for CloudTrail events that had errors:
+
+        errorCode
+
+    To search for errors excluding S3 issues:
+
+        errorCode -s3
+
+    To search for errors excluding S3 that are due to rate limiting:
+
+        errorCode -s3 RequestLimitExceeded
+
+    Technically, the above could be shortened:
+
+        -s3 RequestLimitExceeded
+    """
+    terms = query.split()
+    for term in terms:
+        if term.startswith("-"):
+            events = [e for e in events if not e.contains(term[1:])]
+        else:
+            events = [e for e in events if e.contains(term)]
+    return events
+
+
+def _deep_finder(node, predicate):
+    """Return a generator that yields keys or objects matching `predicate`."""
+    if isinstance(node, dict):
+        for k, v in node.items():
+            if predicate(k):
+                yield k
+            yield from _deep_finder(v, predicate)
+    elif isinstance(node, list):
+        for e in node:
+            yield from _deep_finder(e, predicate)
+    else:
+        if predicate(node):
+            yield node
+
+
 def _colorize(color, string):
     return f"{color}{string}{Style.RESET_ALL}"
 
@@ -651,96 +778,3 @@ TUI_COLOR_THEMES = {
     for k, v in vars(py_cui).items()
     if k.endswith("_ON_BLACK")
 }
-
-
-#############################################################################
-# This code is for deficiencies in the py_cui library. I will actually make a
-# PR for the author of that project at some point.
-#############################################################################
-
-
-class _MyCUI(py_cui.PyCUI):
-    def add_my_scroll_menu(
-        self, title, row, column, row_span=1, column_span=1, padx=1, pady=0
-    ):
-        id = "Widget{}".format(len(self.widgets.keys()))
-        new_scroll_menu = _MyScrollMenu(
-            id, title, self.grid, row, column, row_span, column_span, padx, pady
-        )
-        self.widgets[id] = new_scroll_menu
-        if self.selected_widget is None:
-            self.set_selected_widget(id)
-        return new_scroll_menu
-
-
-class _MyScrollMenu(py_cui.widgets.ScrollMenu):
-    def __init__(
-        self,
-        id,
-        title,
-        grid,
-        row,
-        column,
-        row_span,
-        column_span,
-        padx,
-        pady,
-        to_str=str,
-    ):
-        super().__init__(
-            id, title, grid, row, column, row_span, column_span, padx, pady
-        )
-        self.to_str = to_str
-
-    def page_up(self):
-        shift_by = self.height - (2 * self.pady) - 3
-
-        new_top_view = self.top_view - shift_by
-        new_selected_item = self.selected_item - shift_by
-
-        self.top_view = 0 if new_top_view < 0 else new_top_view
-        self.selected_item = 0 if new_selected_item < 0 else new_selected_item
-
-    def page_down(self):
-        shift_by = self.height - (2 * self.pady) - 3
-        last_item_idx = len(self.view_items) - 1
-
-        new_top_view = self.top_view + shift_by
-        new_selected_item = self.selected_item + shift_by
-
-        self.top_view = (
-            new_top_view - shift_by if new_top_view > last_item_idx else new_top_view
-        )
-        self.selected_item = (
-            last_item_idx if new_selected_item > last_item_idx else new_selected_item
-        )
-
-    def draw(self):
-        super(py_cui.widgets.ScrollMenu, self).draw()
-
-        self.renderer.set_color_mode(
-            self.selected_color if self.selected else self.color
-        )
-        self.renderer.draw_border(self)
-
-        counter = self.pady + 1
-        line_counter = 0
-        for line in (self.to_str(i) for i in self.view_items):
-            if line_counter < self.top_view:
-                line_counter = line_counter + 1
-            else:
-                if counter >= self.height - self.pady - 1:
-                    break
-                if line_counter == self.selected_item:
-                    self.renderer.draw_text(
-                        self, line, self.start_y + counter, selected=True
-                    )
-                else:
-                    self.renderer.draw_text(self, line, self.start_y + counter)
-                counter = counter + 1
-                line_counter = line_counter + 1
-
-        self.renderer.unset_color_mode(
-            self.selected_color if self.selected else self.color
-        )
-        self.renderer.reset_cursor(self)
